<?php
namespace BancoTalentos\Services;

use BancoTalentos\Models\ServiceUser;
use BancoTalentos\Util\GoogleAuth;
use BancoTalentos\Repositories\OperadorRepository;
use BancoTalentos\Services\{ OperadorService, ParamOperService };
use BancoTalentos\Util\BTParameters;
use BancoTalentos\Util\LinkedInAuth;


use BancoTalentos\Util\RequestUtils;
use BancoTalentos\Util\Util;
use BancoTalentos\Util\Utilities;
use DateTime;
use GuzzleHttp\Client;
use Doctrine\ORM\EntityManager;

// Entities
use Teknisa\Model\Gestaopessoas\Entities\BtlParamoper;
use Teknisa\Model\Geral\Entities\Operador;

class BTLoginService {
    private const GOOGLE_SERVICE_NAME = 'google';
    private const LINKEDIN_SERVICE_NAME = 'linkedin';

    private const SERVICES_DATA = [
        self::GOOGLE_SERVICE_NAME => [
            'paramOperColumn' => 'DSGOOGLEID',
            'passwordEnding' => 'e86a8d5528374ad6470ca9f244dcfa22',
        ],
        self::LINKEDIN_SERVICE_NAME => [
            'paramOperColumn' => 'DSLINKEDINID',
            'passwordEnding' => 'a4b4494ea063849c48cf3acc24ced1c5',
        ],
    ];

    /** @var GoogleAuth */
    protected $googleAuth;

    /** @var LinkedInAuth */
    protected $linkedInAuth;

    /** @var OperadorService */
    protected $operadorService;

    /** @var OperadorRepository */
    protected $operadorRepository;

    /** @var string */
    protected $productId;

    /** @var string */
    protected $loginUrl;

    /** @var EntityManager */
    protected $entityManager;

    /** @var Utilities */
    protected $utilities;

    /** @var AnexosService */
    protected $anexosService;

    /** @var BTParameters */
    protected $btParameters;

    /** @var ParamOperService */
    protected $paramOperService;

    public function __construct(
        GoogleAuth $googleAuth,
        OperadorService $operadorService,
        OperadorRepository $operadorRepository,
        string $productId,
        string $projectPath,
        string $loginPath,
        EntityManager $entityManager,
        Utilities $utilities,
        AnexosService $anexosService,
        BTParameters $btParameters,
        LinkedInAuth $linkedInAuth,
        ParamOperService $paramOperService,
        string $projectUrl
    ) {
        $this->googleAuth = $googleAuth;
        $this->operadorService = $operadorService;
        $this->operadorRepository = $operadorRepository;
        $this->productId = $productId;
        $this->loginUrl = $projectUrl . $loginPath;
        $this->entityManager = $entityManager;
        $this->utilities = $utilities;
        $this->anexosService = $anexosService;
        $this->btParameters = $btParameters;
        $this->linkedInAuth = $linkedInAuth;
        $this->paramOperService = $paramOperService;
    }

    public function loginGoogle(
        string $idToken,
        string $requesterUrl
    ) {
        $user = $this->googleAuth->fetchUser($idToken);
        $loginResponse = $this->loginWithService(self::GOOGLE_SERVICE_NAME, $user, $requesterUrl);

        return $loginResponse;
    }

    public function loginLinkedIn(
        string $authorizationCode,
        string $callbackUrl
    ) {
        $user = $this->linkedInAuth->fetchUser(
            $authorizationCode,
            $callbackUrl
        );

        $loginResponse = $this->loginWithService(
            self::LINKEDIN_SERVICE_NAME,
            $user,
            $callbackUrl
        );
        return $loginResponse;
    }

    public function loginWithService(
        string $serviceName,
        ServiceUser $user,
        string $requesterUrl
    ) {
        $operador = $this->getOperador($serviceName, $user);

        if ($user->pictureUrl) {
            $this->setFotoFromUrl($user->pictureUrl, $serviceName);
        }

        $loginResponse = $this->changePasswordAndLogin($serviceName, $operador, $user, $requesterUrl);

        return $loginResponse;
    }

    private function getOperador(
        string $serviceName,
        ServiceUser $user
    ) {
        // é necessário pesquisar primeiro usando o ID pois o usuário
        // pode mudar o e-mail de sua conta google
        $paramOper = $this->getParamOperByUserId($serviceName, $user->id);

        $operador = null;

        if ($paramOper) {
            $operador = $this->getOperadorByParamOper($paramOper);
        } else {
            $operador = $this->operadorService->getOperadorByEmail($user->email);
        }

        $this->entityManager->beginTransaction();

        try {
            if (!$operador) {
                $operador = $this->createOperador($serviceName, $user);
            }

            if (!$paramOper) {
                $this->setParamOper($serviceName, $operador, $user);
            }

            $this->entityManager->commit();
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }

        $nrOrg = $operador->getNrorg();
        $cdOperador = $operador->getCdoperador();
        $nmOperador = $operador->getNmoperador();
        $nrParcNegocio = $operador->getNrparcnegocio();

        $this->btParameters->setNrOrg($nrOrg);
        $this->btParameters->setCdOperador($cdOperador);
        $this->btParameters->setNmOperador($nmOperador);
        $this->btParameters->setNrParcNegocio($nrParcNegocio);

        return $operador;
    }

    private function changePasswordAndLogin(
        string $serviceName,
        Operador $operador,
        ServiceUser $user,
        string $requesterUrl
    ) {
        $nrOrg          = $operador->getNrorg();
        $cdOperador     = $operador->getCdoperador();
        $email          = $operador->getDsemailoper();
        $oldPassword    = $operador->getCdsenhaoperweb();

        $password = $this->getTemporaryPassword($serviceName, $user->id);

        $passwordEncrypt = md5($password);
        $this->changePassword($nrOrg, $cdOperador, $passwordEncrypt);

        try {
            $loginResponse = $this->login($email, $password, $requesterUrl);
        } finally {
            $this->changePassword($nrOrg, $cdOperador, $oldPassword);
        }


        return $loginResponse;
    }

    private function getTemporaryPassword(
        string $serviceName,
        string $id
    ): string {
        $serviceEnding = self::SERVICES_DATA[$serviceName]['passwordEnding'];
        $temporartPassword = $id . $serviceEnding;

        return $temporartPassword;
    }

    private function createOperador(
        string $serviceName,
        ServiceUser $user
    ) {
        $password = $this->getTemporaryPassword($serviceName, $user->id);

        $row = [
            'NMOPERADOR' => $user->name,
            'DSEMAILOPER' => $user->email,
            'PASSWORD' => $password,
            'PASSWORD_CONFIRMATION' => $password,
        ];

        $cdOperador = $this->operadorService->setOperadorAcesso($row);

        $findRow = [
            'CDOPERADOR' => $cdOperador,
            'NRORG' => Util::DEFAULT_ORG,
        ];

        $operador = $this->operadorService->find($findRow);

        return $operador;
    }

    public function changePassword(
        int $nrOrg,
        string $cdOperador,
        string $password
    ): void {
        $params = [
            'P_NRORG' => $nrOrg,
            'P_CDOPERADOR' => $cdOperador,
            'P_SENHA' => $password,
        ];

        $this->operadorRepository->changePassword($params);
    }

    public function login(
        string $email,
        string $password,
        string $requesterUrl
    ): array {
        $route = $this->loginUrl . '/login';

        $params = [
            'requestType' => 'FilterData',
            'filter' => [
                [
                    'name' => 'EMAIL',
                    'operator' => '=',
                    'value' => $email,
                ],
                [
                    'name' => 'PASSWORD',
                    'operator' => '=',
                    'value' => $password,
                ],
                [
                    'name' => 'PRODUCT_ID',
                    'operator' => '=',
                    'value' => $this->productId,
                ],
                [
                    'name' => 'REQUESTER_URL',
                    'operator' => '=',
                    'value' => $requesterUrl,
                ],
                [
                    'name' => 'ATTEMPTS',
                    'operator' => '=',
                    'value' => 1,
                ],
                [
                    'name' => 'KEEP_CONNECTED',
                    'operator' => '=',
                    'value' => 'S',
                ],
            ],
        ];

        $http = new Client([
            'base_uri' => $route
        ]);

        $response = $http->request('POST', $route, [
            'json' => $params,
            'headers' => [
                'Content-Type: application/json',
                'Content-Length: ' . strlen(json_encode($params)),
            ],
        ]);

        $content = $response->getBody()->getContents();
        $responseData = json_decode($content, true);

        if ($responseData['dataset']['userData'] ?? false) {
            $cookiePHP = $response->getHeaders()['Set-Cookie'][0];
            header('Set-Cookie: ' . $cookiePHP);
        }

        return $responseData;
    }

    private function getParamOperByUserId(
        string $serviceName,
        string $value
    ) {
        $column = self::SERVICES_DATA[$serviceName]['paramOperColumn'];

        $row = [
            $column => $value,
        ];

        $paramOper = $this->paramOperService->find($row, [$column]);

        return $paramOper;
    }

    public function setFotoFromUrl(
        string $pictureUrl,
        $service = 'google'
    ): void {
        $currentFoto = $this->anexosService->getFoto();

        if ($currentFoto) {
            return;
        }

        $nrParcNegocio = (int) $this->btParameters->getNrParcNegocio();
        $date = date('Y-m-d');

        $file = $this->utilities->downloadFile($pictureUrl);
        $type = $file['type'];
        $base64 = $this->utilities->getCompleteBase64($file['base64'], $type);

        // foto_google_12345678_2021-10-19.jpg
        $nmAnexo = "foto_${service}_${nrParcNegocio}_${date}.png";

        $row = [
            'NMANEXO' => $nmAnexo,
            'FILE' => $base64,
        ];

        $this->anexosService->setAnexo($row, true);
    }

    private function getOperadorByParamOper(
        BtlParamoper $paramOper
    ) {
        $row = [
            'NRORG' => $paramOper->getNrorg(),
            'CDOPERADOR' => $paramOper->getCdoperador(),
        ];

        $operador = $this->operadorService->find($row);

        return $operador;
    }

    private function setParamOper(
        string $serviceName,
        Operador $operador,
        ServiceUser $user
    ): void {
        $column = self::SERVICES_DATA[$serviceName]['paramOperColumn'];

        $nrOrg = $operador->getNrorg();
        $cdOperador = $operador->getCdoperador();

        $findRow = [
            'NRORG' => $nrOrg,
            'CDOPERADOR' => $cdOperador,
        ];

        $paramOper = $this->paramOperService->find($findRow);

        $row = [
            'NRORG' => $nrOrg,
            'CDOPERADOR' => $cdOperador,
            $column => $user->id,
        ];

        if ($paramOper) {
            $row['DTULTATU'] = new DateTime();
            $row['NRORGULTATU'] = $nrOrg;
            $row['CDOPERULTATU'] = $cdOperador;

            $this->paramOperService->update($row);
            return;
        }

        $row['DTINCLUSAO'] = new DateTime();
        $row['NRORGINCLUSAO'] = $nrOrg;
        $row['CDOPERINCLUSAO'] = $cdOperador;

        $this->paramOperService->set($row);
    }
}